@@ -10,9 +10,11 @@ import sys
 import uuid
 from collections import deque
 from contextlib import closing
+from typing import Any, List, Optional
 from unittest import mock
 
 import pytest
+from aiohappyeyeballs import AddrInfoType
 from yarl import URL
 
 import aiohttp
@@ -539,7 +541,9 @@ async def test__drop_acquire_per_host3(loop) -> None:
     assert conn._acquired_per_host[123] == {789}
 
 
-async def test_tcp_connector_certificate_error(loop) -> None:
+async def test_tcp_connector_certificate_error(
+    loop: Any, start_connection: mock.AsyncMock
+) -> None:
     req = ClientRequest("GET", URL("https://127.0.0.1:443"), loop=loop)
 
     async def certificate_error(*args, **kwargs):
@@ -556,8 +560,10 @@ async def test_tcp_connector_certificate_error(loop) -> None:
     assert isinstance(ctx.value, aiohttp.ClientSSLError)
 
 
-async def test_tcp_connector_server_hostname_default(loop) -> None:
-    conn = aiohttp.TCPConnector(loop=loop)
+async def test_tcp_connector_server_hostname_default(
+    loop: Any, start_connection: mock.AsyncMock
+) -> None:
+    conn = aiohttp.TCPConnector()
 
     with mock.patch.object(
         conn._loop, "create_connection", autospec=True, spec_set=True
@@ -570,8 +576,10 @@ async def test_tcp_connector_server_hostname_default(loop) -> None:
             assert create_connection.call_args.kwargs["server_hostname"] == "127.0.0.1"
 
 
-async def test_tcp_connector_server_hostname_override(loop) -> None:
-    conn = aiohttp.TCPConnector(loop=loop)
+async def test_tcp_connector_server_hostname_override(
+    loop: Any, start_connection: mock.AsyncMock
+) -> None:
+    conn = aiohttp.TCPConnector()
 
     with mock.patch.object(
         conn._loop, "create_connection", autospec=True, spec_set=True
@@ -595,6 +603,7 @@ async def test_tcp_connector_multiple_hosts_errors(loop) -> None:
     ip4 = "192.168.1.4"
     ip5 = "192.168.1.5"
     ips = [ip1, ip2, ip3, ip4, ip5]
+    addrs_tried = []
     ips_tried = []
 
     fingerprint = hashlib.sha256(b"foo").digest()
@@ -624,11 +633,24 @@ async def test_tcp_connector_multiple_hosts_errors(loop) -> None:
     os_error = certificate_error = ssl_error = fingerprint_error = False
     connected = False
 
+    async def start_connection(*args, **kwargs):
+        addr_infos: List[AddrInfoType] = kwargs["addr_infos"]
+
+        first_addr_info = addr_infos[0]
+        first_addr_info_addr = first_addr_info[-1]
+        addrs_tried.append(first_addr_info_addr)
+
+        mock_socket = mock.create_autospec(socket.socket, spec_set=True, instance=True)
+        mock_socket.getpeername.return_value = first_addr_info_addr
+        return mock_socket
+
     async def create_connection(*args, **kwargs):
         nonlocal os_error, certificate_error, ssl_error, fingerprint_error
         nonlocal connected
 
-        ip = args[1]
+        sock = kwargs["sock"]
+        addr_info = sock.getpeername()
+        ip = addr_info[0]
 
         ips_tried.append(ip)
 
@@ -645,6 +667,12 @@ async def test_tcp_connector_multiple_hosts_errors(loop) -> None:
             raise ssl.SSLError
 
         if ip == ip4:
+            sock: socket.socket = kwargs["sock"]
+
+            # Close the socket since we are not actually connecting
+            # and we don't want to leak it.
+            sock.close()
+
             fingerprint_error = True
             tr, pr = mock.Mock(), mock.Mock()
 
@@ -660,12 +688,21 @@ async def test_tcp_connector_multiple_hosts_errors(loop) -> None:
                 if param == "peername":
                     return ("192.168.1.5", 12345)
 
+                if param == "socket":
+                    return sock
+
                 assert False, param
 
             tr.get_extra_info = get_extra_info
             return tr, pr
 
         if ip == ip5:
+            sock: socket.socket = kwargs["sock"]
+
+            # Close the socket since we are not actually connecting
+            # and we don't want to leak it.
+            sock.close()
+
             connected = True
             tr, pr = mock.Mock(), mock.Mock()
 
@@ -687,8 +724,13 @@ async def test_tcp_connector_multiple_hosts_errors(loop) -> None:
 
     conn._loop.create_connection = create_connection
 
-    established_connection = await conn.connect(req, [], ClientTimeout())
-    assert ips == ips_tried
+    with mock.patch(
+        "aiohttp.connector.aiohappyeyeballs.start_connection", start_connection
+    ):
+        established_connection = await conn.connect(req, [], ClientTimeout())
+
+    assert ips_tried == ips
+    assert addrs_tried == [(ip, 443) for ip in ips]
 
     assert os_error
     assert certificate_error
@@ -699,8 +741,214 @@ async def test_tcp_connector_multiple_hosts_errors(loop) -> None:
     established_connection.close()
 
 
-async def test_tcp_connector_resolve_host(loop) -> None:
-    conn = aiohttp.TCPConnector(loop=loop, use_dns_cache=True)
+@pytest.mark.parametrize(
+    ("happy_eyeballs_delay"),
+    [0.1, 0.25, None],
+)
+async def test_tcp_connector_happy_eyeballs(
+    loop: Any, happy_eyeballs_delay: Optional[float]
+) -> None:
+    conn = aiohttp.TCPConnector(happy_eyeballs_delay=happy_eyeballs_delay)
+
+    ip1 = "dead::beef::"
+    ip2 = "192.168.1.1"
+    ips = [ip1, ip2]
+    addrs_tried = []
+
+    req = ClientRequest(
+        "GET",
+        URL("https://mocked.host"),
+        loop=loop,
+    )
+
+    async def _resolve_host(host, port, traces=None):
+        return [
+            {
+                "hostname": host,
+                "host": ip,
+                "port": port,
+                "family": socket.AF_INET6 if ":" in ip else socket.AF_INET,
+                "proto": 0,
+                "flags": socket.AI_NUMERICHOST,
+            }
+            for ip in ips
+        ]
+
+    conn._resolve_host = _resolve_host
+
+    os_error = False
+    connected = False
+
+    async def sock_connect(*args, **kwargs):
+        addr = args[1]
+        nonlocal os_error
+
+        addrs_tried.append(addr)
+
+        if addr[0] == ip1:
+            os_error = True
+            raise OSError
+
+    async def create_connection(*args, **kwargs):
+        sock: socket.socket = kwargs["sock"]
+
+        # Close the socket since we are not actually connecting
+        # and we don't want to leak it.
+        sock.close()
+
+        nonlocal connected
+        connected = True
+        tr = create_mocked_conn(loop)
+        pr = create_mocked_conn(loop)
+        return tr, pr
+
+    conn._loop.sock_connect = sock_connect
+    conn._loop.create_connection = create_connection
+
+    established_connection = await conn.connect(req, [], ClientTimeout())
+
+    assert addrs_tried == [(ip1, 443, 0, 0), (ip2, 443)]
+
+    assert os_error
+    assert connected
+
+    established_connection.close()
+
+
+async def test_tcp_connector_interleave(loop: Any) -> None:
+    conn = aiohttp.TCPConnector(interleave=2)
+
+    ip1 = "192.168.1.1"
+    ip2 = "192.168.1.2"
+    ip3 = "dead::beef::"
+    ip4 = "aaaa::beef::"
+    ip5 = "192.168.1.5"
+    ips = [ip1, ip2, ip3, ip4, ip5]
+    success_ips = []
+    interleave = None
+
+    req = ClientRequest(
+        "GET",
+        URL("https://mocked.host"),
+        loop=loop,
+    )
+
+    async def _resolve_host(host, port, traces=None):
+        return [
+            {
+                "hostname": host,
+                "host": ip,
+                "port": port,
+                "family": socket.AF_INET6 if ":" in ip else socket.AF_INET,
+                "proto": 0,
+                "flags": socket.AI_NUMERICHOST,
+            }
+            for ip in ips
+        ]
+
+    conn._resolve_host = _resolve_host
+
+    async def start_connection(*args, **kwargs):
+        nonlocal interleave
+        addr_infos: List[AddrInfoType] = kwargs["addr_infos"]
+        interleave = kwargs["interleave"]
+        # Mock the 4th host connecting successfully
+        fourth_addr_info = addr_infos[3]
+        fourth_addr_info_addr = fourth_addr_info[-1]
+        mock_socket = mock.create_autospec(socket.socket, spec_set=True, instance=True)
+        mock_socket.getpeername.return_value = fourth_addr_info_addr
+        return mock_socket
+
+    async def create_connection(*args, **kwargs):
+        sock = kwargs["sock"]
+        addr_info = sock.getpeername()
+        ip = addr_info[0]
+
+        success_ips.append(ip)
+
+        sock: socket.socket = kwargs["sock"]
+        # Close the socket since we are not actually connecting
+        # and we don't want to leak it.
+        sock.close()
+        tr = create_mocked_conn(loop)
+        pr = create_mocked_conn(loop)
+        return tr, pr
+
+    conn._loop.create_connection = create_connection
+
+    with mock.patch(
+        "aiohttp.connector.aiohappyeyeballs.start_connection", start_connection
+    ):
+        established_connection = await conn.connect(req, [], ClientTimeout())
+
+    assert success_ips == [ip4]
+    assert interleave == 2
+    established_connection.close()
+
+
+async def test_tcp_connector_family_is_respected(loop: Any) -> None:
+    conn = aiohttp.TCPConnector(family=socket.AF_INET)
+
+    ip1 = "dead::beef::"
+    ip2 = "192.168.1.1"
+    ips = [ip1, ip2]
+    addrs_tried = []
+
+    req = ClientRequest(
+        "GET",
+        URL("https://mocked.host"),
+        loop=loop,
+    )
+
+    async def _resolve_host(host, port, traces=None):
+        return [
+            {
+                "hostname": host,
+                "host": ip,
+                "port": port,
+                "family": socket.AF_INET6 if ":" in ip else socket.AF_INET,
+                "proto": 0,
+                "flags": socket.AI_NUMERICHOST,
+            }
+            for ip in ips
+        ]
+
+    conn._resolve_host = _resolve_host
+    connected = False
+
+    async def sock_connect(*args, **kwargs):
+        addr = args[1]
+        addrs_tried.append(addr)
+
+    async def create_connection(*args, **kwargs):
+        sock: socket.socket = kwargs["sock"]
+
+        # Close the socket since we are not actually connecting
+        # and we don't want to leak it.
+        sock.close()
+
+        nonlocal connected
+        connected = True
+        tr = create_mocked_conn(loop)
+        pr = create_mocked_conn(loop)
+        return tr, pr
+
+    conn._loop.sock_connect = sock_connect
+    conn._loop.create_connection = create_connection
+
+    established_connection = await conn.connect(req, [], ClientTimeout())
+
+    # We should only try the IPv4 address since we specified
+    # the family to be AF_INET
+    assert addrs_tried == [(ip2, 443)]
+
+    assert connected
+
+    established_connection.close()
+
+
+async def test_tcp_connector_resolve_host(loop: Any) -> None:
+    conn = aiohttp.TCPConnector(use_dns_cache=True)
 
     res = await conn._resolve_host("localhost", 8080)
     assert res
@@ -2047,7 +2295,8 @@ async def test_tcp_connector_raise_connector_ssl_error(
     session = aiohttp.ClientSession(connector=conn)
     url = srv.make_url("/")
 
-    with pytest.raises(aiohttp.ClientConnectorCertificateError) as ctx:
+    err = aiohttp.ClientConnectorCertificateError
+    with pytest.raises(err) as ctx:
         await session.get(url)
 
     assert isinstance(ctx.value, aiohttp.ClientConnectorCertificateError)
