@@ -8,38 +8,35 @@ import socket
 import ssl
 import subprocess
 import sys
+import time
+from typing import Callable, NoReturn, Set
 from unittest import mock
 from uuid import uuid4
 
 import pytest
 
-from aiohttp import web
-from aiohttp.helpers import PY_37
+from aiohttp import ClientConnectorError, ClientSession, ClientTimeout, WSCloseCode, web
 from aiohttp.test_utils import make_mocked_coro
+from aiohttp.web_runner import BaseRunner
 
 # Test for features of OS' socket support
-_has_unix_domain_socks = hasattr(socket, "AF_UNIX")
-if _has_unix_domain_socks:
-    _abstract_path_sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
-    try:
-        _abstract_path_sock.bind(b"\x00" + uuid4().hex.encode("ascii"))  # type: ignore  # noqa
-    except FileNotFoundError:
-        _abstract_path_failed = True
-    else:
-        _abstract_path_failed = False
-    finally:
-        _abstract_path_sock.close()
-        del _abstract_path_sock
+if hasattr(socket, "AF_UNIX"):
+    with socket.socket(socket.AF_UNIX, socket.SOCK_STREAM) as _abstract_path_sock:
+        try:
+            _abstract_path_sock.bind(b"\x00" + uuid4().hex.encode("ascii"))
+        except FileNotFoundError:
+            _abstract_path_failed = True
+        else:
+            _abstract_path_failed = False
+        finally:
+            del _abstract_path_sock
 else:
     _abstract_path_failed = True
 
 skip_if_no_abstract_paths = pytest.mark.skipif(
     _abstract_path_failed, reason="Linux-style abstract paths are not supported."
 )
-skip_if_no_unix_socks = pytest.mark.skipif(
-    not _has_unix_domain_socks, reason="Unix domain sockets are not supported"
-)
-del _has_unix_domain_socks, _abstract_path_failed
+del _abstract_path_failed
 
 HAS_IPV6 = socket.has_ipv6
 if HAS_IPV6:
@@ -47,17 +44,12 @@ if HAS_IPV6:
     # support, but the target system still may not have it.
     # So let's ensure that we really have IPv6 support.
     try:
-        socket.socket(socket.AF_INET6, socket.SOCK_STREAM)
+        with socket.socket(socket.AF_INET6, socket.SOCK_STREAM):
+            pass
     except OSError:
         HAS_IPV6 = False
 
 
-# tokio event loop does not allow to override attributes
-def skip_if_no_dict(loop):
-    if not hasattr(loop, "__dict__"):
-        pytest.skip("can not override loop attributes")
-
-
 def skip_if_on_windows():
     if platform.system() == "Windows":
         pytest.skip("the test is not valid for Windows")
@@ -65,7 +57,6 @@ def skip_if_on_windows():
 
 @pytest.fixture
 def patched_loop(loop):
-    skip_if_no_dict(loop)
     server = mock.Mock()
     server.wait_closed = make_mocked_coro(None)
     loop.create_server = make_mocked_coro(server)
@@ -93,7 +84,7 @@ def test_run_app_http(patched_loop) -> None:
     cleanup_handler = make_mocked_coro()
     app.on_cleanup.append(cleanup_handler)
 
-    web.run_app(app, print=stopper(patched_loop))
+    web.run_app(app, print=stopper(patched_loop), loop=patched_loop)
 
     patched_loop.create_server.assert_called_with(
         mock.ANY, None, 8080, ssl=None, backlog=128, reuse_address=None, reuse_port=None
@@ -104,7 +95,7 @@ def test_run_app_http(patched_loop) -> None:
 
 def test_run_app_close_loop(patched_loop) -> None:
     app = web.Application()
-    web.run_app(app, print=stopper(patched_loop))
+    web.run_app(app, print=stopper(patched_loop), loop=patched_loop)
 
     patched_loop.create_server.assert_called_with(
         mock.ANY, None, 8080, ssl=None, backlog=128, reuse_address=None, reuse_port=None
@@ -424,7 +415,7 @@ def test_run_app_mixed_bindings(
     run_app_kwargs, expected_server_calls, expected_unix_server_calls, patched_loop
 ):
     app = web.Application()
-    web.run_app(app, print=stopper(patched_loop), **run_app_kwargs)
+    web.run_app(app, print=stopper(patched_loop), **run_app_kwargs, loop=patched_loop)
 
     assert patched_loop.create_unix_server.mock_calls == expected_unix_server_calls
     assert patched_loop.create_server.mock_calls == expected_server_calls
@@ -434,7 +425,9 @@ def test_run_app_https(patched_loop) -> None:
     app = web.Application()
 
     ssl_context = ssl.create_default_context()
-    web.run_app(app, ssl_context=ssl_context, print=stopper(patched_loop))
+    web.run_app(
+        app, ssl_context=ssl_context, print=stopper(patched_loop), loop=patched_loop
+    )
 
     patched_loop.create_server.assert_called_with(
         mock.ANY,
@@ -452,7 +445,9 @@ def test_run_app_nondefault_host_port(patched_loop, aiohttp_unused_port) -> None
     host = "127.0.0.1"
 
     app = web.Application()
-    web.run_app(app, host=host, port=port, print=stopper(patched_loop))
+    web.run_app(
+        app, host=host, port=port, print=stopper(patched_loop), loop=patched_loop
+    )
 
     patched_loop.create_server.assert_called_with(
         mock.ANY, host, port, ssl=None, backlog=128, reuse_address=None, reuse_port=None
@@ -463,7 +458,7 @@ def test_run_app_multiple_hosts(patched_loop) -> None:
     hosts = ("127.0.0.1", "127.0.0.2")
 
     app = web.Application()
-    web.run_app(app, host=hosts, print=stopper(patched_loop))
+    web.run_app(app, host=hosts, print=stopper(patched_loop), loop=patched_loop)
 
     calls = map(
         lambda h: mock.call(
@@ -482,7 +477,7 @@ def test_run_app_multiple_hosts(patched_loop) -> None:
 
 def test_run_app_custom_backlog(patched_loop) -> None:
     app = web.Application()
-    web.run_app(app, backlog=10, print=stopper(patched_loop))
+    web.run_app(app, backlog=10, print=stopper(patched_loop), loop=patched_loop)
 
     patched_loop.create_server.assert_called_with(
         mock.ANY, None, 8080, ssl=None, backlog=10, reuse_address=None, reuse_port=None
@@ -491,49 +486,60 @@ def test_run_app_custom_backlog(patched_loop) -> None:
 
 def test_run_app_custom_backlog_unix(patched_loop) -> None:
     app = web.Application()
-    web.run_app(app, path="/tmp/tmpsock.sock", backlog=10, print=stopper(patched_loop))
+    web.run_app(
+        app,
+        path="/tmp/tmpsock.sock",
+        backlog=10,
+        print=stopper(patched_loop),
+        loop=patched_loop,
+    )
 
     patched_loop.create_unix_server.assert_called_with(
         mock.ANY, "/tmp/tmpsock.sock", ssl=None, backlog=10
     )
 
 
-@skip_if_no_unix_socks
-def test_run_app_http_unix_socket(patched_loop, shorttmpdir) -> None:
+def test_run_app_http_unix_socket(patched_loop, unix_sockname) -> None:
     app = web.Application()
 
-    sock_path = str(shorttmpdir / "socket.sock")
     printer = mock.Mock(wraps=stopper(patched_loop))
-    web.run_app(app, path=sock_path, print=printer)
+    web.run_app(app, path=unix_sockname, print=printer, loop=patched_loop)
 
     patched_loop.create_unix_server.assert_called_with(
-        mock.ANY, sock_path, ssl=None, backlog=128
+        mock.ANY, unix_sockname, ssl=None, backlog=128
     )
-    assert f"http://unix:{sock_path}:" in printer.call_args[0][0]
+    assert f"http://unix:{unix_sockname}:" in printer.call_args[0][0]
 
 
-@skip_if_no_unix_socks
-def test_run_app_https_unix_socket(patched_loop, shorttmpdir) -> None:
+def test_run_app_https_unix_socket(patched_loop, unix_sockname) -> None:
     app = web.Application()
 
-    sock_path = str(shorttmpdir / "socket.sock")
     ssl_context = ssl.create_default_context()
     printer = mock.Mock(wraps=stopper(patched_loop))
-    web.run_app(app, path=sock_path, ssl_context=ssl_context, print=printer)
+    web.run_app(
+        app,
+        path=unix_sockname,
+        ssl_context=ssl_context,
+        print=printer,
+        loop=patched_loop,
+    )
 
     patched_loop.create_unix_server.assert_called_with(
-        mock.ANY, sock_path, ssl=ssl_context, backlog=128
+        mock.ANY, unix_sockname, ssl=ssl_context, backlog=128
     )
-    assert f"https://unix:{sock_path}:" in printer.call_args[0][0]
+    assert f"https://unix:{unix_sockname}:" in printer.call_args[0][0]
 
 
-@skip_if_no_unix_socks
+@pytest.mark.skipif(not hasattr(socket, "AF_UNIX"), reason="requires UNIX sockets")
 @skip_if_no_abstract_paths
 def test_run_app_abstract_linux_socket(patched_loop) -> None:
     sock_path = b"\x00" + uuid4().hex.encode("ascii")
     app = web.Application()
     web.run_app(
-        app, path=sock_path.decode("ascii", "ignore"), print=stopper(patched_loop)
+        app,
+        path=sock_path.decode("ascii", "ignore"),
+        print=stopper(patched_loop),
+        loop=patched_loop,
     )
 
     patched_loop.create_unix_server.assert_called_with(
@@ -546,16 +552,16 @@ def test_run_app_preexisting_inet_socket(patched_loop, mocker) -> None:
 
     sock = socket.socket()
     with contextlib.closing(sock):
-        sock.bind(("0.0.0.0", 0))
+        sock.bind(("127.0.0.1", 0))
         _, port = sock.getsockname()
 
         printer = mock.Mock(wraps=stopper(patched_loop))
-        web.run_app(app, sock=sock, print=printer)
+        web.run_app(app, sock=sock, print=printer, loop=patched_loop)
 
         patched_loop.create_server.assert_called_with(
             mock.ANY, sock=sock, backlog=128, ssl=None
         )
-        assert f"http://0.0.0.0:{port}" in printer.call_args[0][0]
+        assert f"http://127.0.0.1:{port}" in printer.call_args[0][0]
 
 
 @pytest.mark.skipif(not HAS_IPV6, reason="IPv6 is not available")
@@ -564,19 +570,19 @@ def test_run_app_preexisting_inet6_socket(patched_loop) -> None:
 
     sock = socket.socket(socket.AF_INET6)
     with contextlib.closing(sock):
-        sock.bind(("::", 0))
+        sock.bind(("::1", 0))
         port = sock.getsockname()[1]
 
         printer = mock.Mock(wraps=stopper(patched_loop))
-        web.run_app(app, sock=sock, print=printer)
+        web.run_app(app, sock=sock, print=printer, loop=patched_loop)
 
         patched_loop.create_server.assert_called_with(
             mock.ANY, sock=sock, backlog=128, ssl=None
         )
-        assert f"http://[::]:{port}" in printer.call_args[0][0]
+        assert f"http://[::1]:{port}" in printer.call_args[0][0]
 
 
-@skip_if_no_unix_socks
+@pytest.mark.skipif(not hasattr(socket, "AF_UNIX"), reason="requires UNIX sockets")
 def test_run_app_preexisting_unix_socket(patched_loop, mocker) -> None:
     app = web.Application()
 
@@ -587,7 +593,7 @@ def test_run_app_preexisting_unix_socket(patched_loop, mocker) -> None:
         os.unlink(sock_path)
 
         printer = mock.Mock(wraps=stopper(patched_loop))
-        web.run_app(app, sock=sock, print=printer)
+        web.run_app(app, sock=sock, print=printer, loop=patched_loop)
 
         patched_loop.create_server.assert_called_with(
             mock.ANY, sock=sock, backlog=128, ssl=None
@@ -601,13 +607,13 @@ def test_run_app_multiple_preexisting_sockets(patched_loop) -> None:
     sock1 = socket.socket()
     sock2 = socket.socket()
     with contextlib.closing(sock1), contextlib.closing(sock2):
-        sock1.bind(("0.0.0.0", 0))
+        sock1.bind(("localhost", 0))
         _, port1 = sock1.getsockname()
-        sock2.bind(("0.0.0.0", 0))
+        sock2.bind(("localhost", 0))
         _, port2 = sock2.getsockname()
 
         printer = mock.Mock(wraps=stopper(patched_loop))
-        web.run_app(app, sock=(sock1, sock2), print=printer)
+        web.run_app(app, sock=(sock1, sock2), print=printer, loop=patched_loop)
 
         patched_loop.create_server.assert_has_calls(
             [
@@ -615,8 +621,8 @@ def test_run_app_multiple_preexisting_sockets(patched_loop) -> None:
                 mock.call(mock.ANY, sock=sock2, backlog=128, ssl=None),
             ]
         )
-        assert f"http://0.0.0.0:{port1}" in printer.call_args[0][0]
-        assert f"http://0.0.0.0:{port2}" in printer.call_args[0][0]
+        assert f"http://127.0.0.1:{port1}" in printer.call_args[0][0]
+        assert f"http://127.0.0.1:{port2}" in printer.call_args[0][0]
 
 
 _script_test_signal = """
@@ -630,27 +636,29 @@ web.run_app(app, host=())
 def test_sigint() -> None:
     skip_if_on_windows()
 
-    proc = subprocess.Popen(
-        [sys.executable, "-u", "-c", _script_test_signal], stdout=subprocess.PIPE
-    )
-    for line in proc.stdout:
-        if line.startswith(b"======== Running on"):
-            break
-    proc.send_signal(signal.SIGINT)
-    assert proc.wait() == 0
+    with subprocess.Popen(
+        [sys.executable, "-u", "-c", _script_test_signal],
+        stdout=subprocess.PIPE,
+    ) as proc:
+        for line in proc.stdout:
+            if line.startswith(b"======== Running on"):
+                break
+        proc.send_signal(signal.SIGINT)
+        assert proc.wait() == 0
 
 
 def test_sigterm() -> None:
     skip_if_on_windows()
 
-    proc = subprocess.Popen(
-        [sys.executable, "-u", "-c", _script_test_signal], stdout=subprocess.PIPE
-    )
-    for line in proc.stdout:
-        if line.startswith(b"======== Running on"):
-            break
-    proc.terminate()
-    assert proc.wait() == 0
+    with subprocess.Popen(
+        [sys.executable, "-u", "-c", _script_test_signal],
+        stdout=subprocess.PIPE,
+    ) as proc:
+        for line in proc.stdout:
+            if line.startswith(b"======== Running on"):
+                break
+        proc.terminate()
+        assert proc.wait() == 0
 
 
 def test_startup_cleanup_signals_even_on_failure(patched_loop) -> None:
@@ -663,7 +671,7 @@ def test_startup_cleanup_signals_even_on_failure(patched_loop) -> None:
     app.on_cleanup.append(cleanup_handler)
 
     with pytest.raises(RuntimeError):
-        web.run_app(app, print=stopper(patched_loop))
+        web.run_app(app, print=stopper(patched_loop), loop=patched_loop)
 
     startup_handler.assert_called_once_with(app)
     cleanup_handler.assert_called_once_with(app)
@@ -681,7 +689,7 @@ def test_run_app_coro(patched_loop) -> None:
         app.on_cleanup.append(cleanup_handler)
         return app
 
-    web.run_app(make_app(), print=stopper(patched_loop))
+    web.run_app(make_app(), print=stopper(patched_loop), loop=patched_loop)
 
     patched_loop.create_server.assert_called_with(
         mock.ANY, None, 8080, ssl=None, backlog=128, reuse_address=None, reuse_port=None
@@ -702,7 +710,12 @@ def test_run_app_default_logger(monkeypatch, patched_loop):
     mock_logger.configure_mock(**attrs)
 
     app = web.Application()
-    web.run_app(app, print=stopper(patched_loop), access_log=mock_logger)
+    web.run_app(
+        app,
+        print=stopper(patched_loop),
+        access_log=mock_logger,
+        loop=patched_loop,
+    )
     mock_logger.setLevel.assert_any_call(logging.DEBUG)
     mock_logger.hasHandlers.assert_called_with()
     assert isinstance(mock_logger.addHandler.call_args[0][0], logging.StreamHandler)
@@ -720,7 +733,12 @@ def test_run_app_default_logger_setup_requires_debug(patched_loop):
     mock_logger.configure_mock(**attrs)
 
     app = web.Application()
-    web.run_app(app, print=stopper(patched_loop), access_log=mock_logger)
+    web.run_app(
+        app,
+        print=stopper(patched_loop),
+        access_log=mock_logger,
+        loop=patched_loop,
+    )
     mock_logger.setLevel.assert_not_called()
     mock_logger.hasHandlers.assert_not_called()
     mock_logger.addHandler.assert_not_called()
@@ -738,7 +756,12 @@ def test_run_app_default_logger_setup_requires_default_logger(patched_loop):
     mock_logger.configure_mock(**attrs)
 
     app = web.Application()
-    web.run_app(app, print=stopper(patched_loop), access_log=mock_logger)
+    web.run_app(
+        app,
+        print=stopper(patched_loop),
+        access_log=mock_logger,
+        loop=patched_loop,
+    )
     mock_logger.setLevel.assert_not_called()
     mock_logger.hasHandlers.assert_not_called()
     mock_logger.addHandler.assert_not_called()
@@ -756,7 +779,12 @@ def test_run_app_default_logger_setup_only_if_unconfigured(patched_loop):
     mock_logger.configure_mock(**attrs)
 
     app = web.Application()
-    web.run_app(app, print=stopper(patched_loop), access_log=mock_logger)
+    web.run_app(
+        app,
+        print=stopper(patched_loop),
+        access_log=mock_logger,
+        loop=patched_loop,
+    )
     mock_logger.setLevel.assert_not_called()
     mock_logger.hasHandlers.assert_called_with()
     mock_logger.addHandler.assert_not_called()
@@ -773,7 +801,7 @@ def test_run_app_cancels_all_pending_tasks(patched_loop):
 
     app.on_startup.append(on_startup)
 
-    web.run_app(app, print=stopper(patched_loop))
+    web.run_app(app, print=stopper(patched_loop), loop=patched_loop)
     assert task.cancelled()
 
 
@@ -791,7 +819,7 @@ def test_run_app_cancels_done_tasks(patched_loop):
 
     app.on_startup.append(on_startup)
 
-    web.run_app(app, print=stopper(patched_loop))
+    web.run_app(app, print=stopper(patched_loop), loop=patched_loop)
     assert task.done()
 
 
@@ -817,7 +845,7 @@ def test_run_app_cancels_failed_tasks(patched_loop):
 
     exc_handler = mock.Mock()
     patched_loop.set_exception_handler(exc_handler)
-    web.run_app(app, print=stopper(patched_loop))
+    web.run_app(app, print=stopper(patched_loop), loop=patched_loop)
     assert task.done()
 
     msg = {
@@ -828,7 +856,24 @@ def test_run_app_cancels_failed_tasks(patched_loop):
     exc_handler.assert_called_with(patched_loop, msg)
 
 
-@pytest.mark.skipif(not PY_37, reason="contextvars support is required")
+def test_run_app_keepalive_timeout(patched_loop, mocker, monkeypatch):
+    new_timeout = 1234
+    base_runner_init_orig = BaseRunner.__init__
+
+    def base_runner_init_spy(self, *args, **kwargs):
+        assert kwargs["keepalive_timeout"] == new_timeout
+        base_runner_init_orig(self, *args, **kwargs)
+
+    app = web.Application()
+    monkeypatch.setattr(BaseRunner, "__init__", base_runner_init_spy)
+    web.run_app(
+        app,
+        keepalive_timeout=new_timeout,
+        print=stopper(patched_loop),
+        loop=patched_loop,
+    )
+
+
 def test_run_app_context_vars(patched_loop):
     from contextvars import ContextVar
 
@@ -857,5 +902,342 @@ def test_run_app_context_vars(patched_loop):
         count += 1
         return app
 
-    web.run_app(init(), print=stopper(patched_loop))
+    web.run_app(init(), print=stopper(patched_loop), loop=patched_loop)
     assert count == 3
+
+
+class TestShutdown:
+    def raiser(self) -> NoReturn:
+        raise KeyboardInterrupt
+
+    async def stop(self, request: web.Request) -> web.Response:
+        asyncio.get_running_loop().call_soon(self.raiser)
+        return web.Response()
+
+    def run_app(self, port: int, timeout: int, task, extra_test=None) -> asyncio.Task:
+        num_connections = -1
+
+        class DictRecordClear(dict):
+            def clear(self):
+                nonlocal num_connections
+                # During Server.shutdown() we want to know how many connections still
+                # remained before it got cleared. If the handler completed successfully
+                # the connection should've been removed already. If not, this may
+                # indicate a memory leak.
+                num_connections = len(self)
+                super().clear()
+
+        class ServerWithRecordClear(web.Server):
+            def __init__(self, *args, **kwargs):
+                super().__init__(*args, **kwargs)
+                self._connections = DictRecordClear()
+
+        async def test() -> None:
+            await asyncio.sleep(0.5)
+            async with ClientSession() as sess:
+                for _ in range(5):  # pragma: no cover
+                    try:
+                        with pytest.raises(asyncio.TimeoutError):
+                            async with sess.get(
+                                f"http://localhost:{port}/",
+                                timeout=ClientTimeout(total=0.1),
+                            ):
+                                pass
+                    except ClientConnectorError:
+                        await asyncio.sleep(0.5)
+                    else:
+                        break
+                async with sess.get(f"http://localhost:{port}/stop"):
+                    pass
+
+                if extra_test:
+                    await extra_test(sess)
+
+        async def run_test(app: web.Application) -> None:
+            nonlocal test_task
+            test_task = asyncio.create_task(test())
+            yield
+            await test_task
+
+        async def handler(request: web.Request) -> web.Response:
+            nonlocal t
+            t = asyncio.create_task(task())
+            await t
+            return web.Response(text="FOO")
+
+        t = test_task = None
+        app = web.Application()
+        app.cleanup_ctx.append(run_test)
+        app.router.add_get("/", handler)
+        app.router.add_get("/stop", self.stop)
+
+        with mock.patch("aiohttp.web_app.Server", ServerWithRecordClear):
+            web.run_app(app, port=port, shutdown_timeout=timeout)
+        assert test_task.exception() is None
+        return t, num_connections
+
+    def test_shutdown_wait_for_handler(
+        self, aiohttp_unused_port: Callable[[], int]
+    ) -> None:
+        port = aiohttp_unused_port()
+        finished = False
+
+        async def task():
+            nonlocal finished
+            await asyncio.sleep(2)
+            finished = True
+
+        t, connection_count = self.run_app(port, 3, task)
+
+        assert finished is True
+        assert t.done()
+        assert not t.cancelled()
+        assert connection_count == 0
+
+    def test_shutdown_timeout_handler(
+        self, aiohttp_unused_port: Callable[[], int]
+    ) -> None:
+        port = aiohttp_unused_port()
+        finished = False
+
+        async def task():
+            nonlocal finished
+            await asyncio.sleep(2)
+            finished = True
+
+        t, connection_count = self.run_app(port, 1, task)
+
+        assert finished is False
+        assert t.done()
+        assert t.cancelled()
+        assert connection_count == 1
+
+    def test_shutdown_timeout_not_reached(
+        self, aiohttp_unused_port: Callable[[], int]
+    ) -> None:
+        port = aiohttp_unused_port()
+        finished = False
+
+        async def task():
+            nonlocal finished
+            await asyncio.sleep(1)
+            finished = True
+
+        start_time = time.time()
+        t, connection_count = self.run_app(port, 15, task)
+
+        assert finished is True
+        assert t.done()
+        assert connection_count == 0
+        # Verify run_app has not waited for timeout.
+        assert time.time() - start_time < 10
+
+    def test_shutdown_new_conn_rejected(
+        self, aiohttp_unused_port: Callable[[], int]
+    ) -> None:
+        port = aiohttp_unused_port()
+        finished = False
+
+        async def task() -> None:
+            nonlocal finished
+            await asyncio.sleep(9)
+            finished = True
+
+        async def test(sess: ClientSession) -> None:
+            # Ensure we are in the middle of shutdown (waiting for task()).
+            await asyncio.sleep(1)
+            with pytest.raises(ClientConnectorError):
+                # Use a new session to try and open a new connection.
+                async with ClientSession() as sess:
+                    async with sess.get(f"http://localhost:{port}/"):
+                        pass
+            assert finished is False
+
+        t, connection_count = self.run_app(port, 10, task, test)
+
+        assert finished is True
+        assert t.done()
+        assert connection_count == 0
+
+    def test_shutdown_pending_handler_responds(
+        self, aiohttp_unused_port: Callable[[], int]
+    ) -> None:
+        port = aiohttp_unused_port()
+        finished = False
+
+        async def test() -> None:
+            async def test_resp(sess):
+                async with sess.get(f"http://localhost:{port}/") as resp:
+                    assert await resp.text() == "FOO"
+
+            await asyncio.sleep(1)
+            async with ClientSession() as sess:
+                t = asyncio.create_task(test_resp(sess))
+                await asyncio.sleep(1)
+                # Handler is in-progress while we trigger server shutdown.
+                async with sess.get(f"http://localhost:{port}/stop"):
+                    pass
+
+                assert finished is False
+                # Handler should still complete and produce a response.
+                await t
+
+        async def run_test(app: web.Application) -> None:
+            nonlocal t
+            t = asyncio.create_task(test())
+            yield
+            await t
+
+        async def handler(request: web.Request) -> web.Response:
+            nonlocal finished
+            await asyncio.sleep(3)
+            finished = True
+            return web.Response(text="FOO")
+
+        t = None
+        app = web.Application()
+        app.cleanup_ctx.append(run_test)
+        app.router.add_get("/", handler)
+        app.router.add_get("/stop", self.stop)
+
+        web.run_app(app, port=port, shutdown_timeout=5)
+        assert t.exception() is None
+        assert finished is True
+
+    def test_shutdown_close_idle_keepalive(
+        self, aiohttp_unused_port: Callable[[], int]
+    ) -> None:
+        port = aiohttp_unused_port()
+
+        async def test() -> None:
+            await asyncio.sleep(1)
+            async with ClientSession() as sess:
+                async with sess.get(f"http://localhost:{port}/stop"):
+                    pass
+
+                # Hold on to keep-alive connection.
+                await asyncio.sleep(5)
+
+        async def run_test(app: web.Application) -> None:
+            nonlocal t
+            t = asyncio.create_task(test())
+            yield
+            t.cancel()
+            with contextlib.suppress(asyncio.CancelledError):
+                await t
+
+        t = None
+        app = web.Application()
+        app.cleanup_ctx.append(run_test)
+        app.router.add_get("/stop", self.stop)
+
+        web.run_app(app, port=port, shutdown_timeout=10)
+        # If connection closed, then test() will be cancelled in cleanup_ctx.
+        # If not, then shutdown_timeout will allow it to sleep until complete.
+        assert t.cancelled()
+
+    def test_shutdown_close_websockets(
+        self, aiohttp_unused_port: Callable[[], int]
+    ) -> None:
+        port = aiohttp_unused_port()
+        WS = web.AppKey("ws", Set[web.WebSocketResponse])
+        client_finished = server_finished = False
+
+        async def ws_handler(request: web.Request) -> web.WebSocketResponse:
+            ws = web.WebSocketResponse()
+            await ws.prepare(request)
+            request.app[WS].add(ws)
+            async for msg in ws:
+                pass
+            nonlocal server_finished
+            server_finished = True
+            return ws
+
+        async def close_websockets(app: web.Application) -> None:
+            for ws in app[WS]:
+                await ws.close(code=WSCloseCode.GOING_AWAY)
+
+        async def test() -> None:
+            await asyncio.sleep(1)
+            async with ClientSession() as sess:
+                async with sess.ws_connect(f"http://localhost:{port}/ws") as ws:
+                    async with sess.get(f"http://localhost:{port}/stop"):
+                        pass
+
+                    async for msg in ws:
+                        pass
+                    nonlocal client_finished
+                    client_finished = True
+
+        async def run_test(app: web.Application) -> None:
+            nonlocal t
+            t = asyncio.create_task(test())
+            yield
+            t.cancel()
+            with contextlib.suppress(asyncio.CancelledError):
+                await t
+
+        t = None
+        app = web.Application()
+        app[WS] = set()
+        app.on_shutdown.append(close_websockets)
+        app.cleanup_ctx.append(run_test)
+        app.router.add_get("/ws", ws_handler)
+        app.router.add_get("/stop", self.stop)
+
+        start = time.time()
+        web.run_app(app, port=port, shutdown_timeout=10)
+        assert time.time() - start < 5
+        assert client_finished
+        assert server_finished
+
+    def test_shutdown_handler_cancellation_suppressed(
+        self, aiohttp_unused_port: Callable[[], int]
+    ) -> None:
+        port = aiohttp_unused_port()
+        actions = []
+
+        async def test() -> None:
+            async def test_resp(sess):
+                t = ClientTimeout(total=0.4)
+                with pytest.raises(asyncio.TimeoutError):
+                    async with sess.get(f"http://localhost:{port}/", timeout=t) as resp:
+                        assert await resp.text() == "FOO"
+                actions.append("CANCELLED")
+
+            async with ClientSession() as sess:
+                t = asyncio.create_task(test_resp(sess))
+                await asyncio.sleep(0.5)
+                # Handler is in-progress while we trigger server shutdown.
+                actions.append("PRESTOP")
+                async with sess.get(f"http://localhost:{port}/stop"):
+                    pass
+
+                actions.append("STOPPING")
+                # Handler should still complete and produce a response.
+                await t
+
+        async def run_test(app: web.Application) -> None:
+            nonlocal t
+            t = asyncio.create_task(test())
+            yield
+            await t
+
+        async def handler(request: web.Request) -> web.Response:
+            try:
+                await asyncio.sleep(5)
+            except asyncio.CancelledError:
+                actions.append("SUPPRESSED")
+                await asyncio.sleep(2)
+                actions.append("DONE")
+            return web.Response(text="FOO")
+
+        t = None
+        app = web.Application()
+        app.cleanup_ctx.append(run_test)
+        app.router.add_get("/", handler)
+        app.router.add_get("/stop", self.stop)
+
+        web.run_app(app, port=port, shutdown_timeout=2, handler_cancellation=True)
+        assert t.exception() is None
+        assert actions == ["CANCELLED", "SUPPRESSED", "PRESTOP", "STOPPING", "DONE"]
