@@ -1,20 +1,52 @@
 import asyncio
+import bz2
+import gzip
 import pathlib
 import socket
 import zlib
-from typing import Any, Iterable
+from typing import Any, Iterable, Optional
 
 import pytest
 
 import aiohttp
 from aiohttp import web
 
+try:
+    import brotlicffi as brotli
+except ImportError:
+    import brotli
+
 try:
     import ssl
 except ImportError:
     ssl = None  # type: ignore
 
 
+HELLO_AIOHTTP = b"Hello aiohttp! :-)\n"
+
+
+@pytest.fixture(scope="module")
+def hello_txt(request, tmp_path_factory) -> pathlib.Path:
+    """Create a temp path with hello.txt and compressed versions.
+
+    The uncompressed text file path is returned by default. Alternatively, an
+    indirect parameter can be passed with an encoding to get a compressed path.
+    """
+    txt = tmp_path_factory.mktemp("hello-") / "hello.txt"
+    hello = {
+        None: txt,
+        "gzip": txt.with_suffix(f"{txt.suffix}.gz"),
+        "br": txt.with_suffix(f"{txt.suffix}.br"),
+        "bzip2": txt.with_suffix(f"{txt.suffix}.bz2"),
+    }
+    # Uncompressed file is not actually written to test it is not required.
+    hello["gzip"].write_bytes(gzip.compress(HELLO_AIOHTTP))
+    hello["br"].write_bytes(brotli.compress(HELLO_AIOHTTP))
+    hello["bzip2"].write_bytes(bz2.compress(HELLO_AIOHTTP))
+    encoding = getattr(request, "param", None)
+    return hello[encoding]
+
+
 @pytest.fixture
 def loop_without_sendfile(loop):
     def sendfile(*args, **kwargs):
@@ -201,100 +233,123 @@ async def test_static_file_with_content_type(aiohttp_client, sender) -> None:
     await client.close()
 
 
-async def test_static_file_custom_content_type(aiohttp_client, sender) -> None:
-    filepath = pathlib.Path(__file__).parent / "hello.txt.gz"
-
-    async def handler(request):
-        resp = sender(filepath, chunk_size=16)
-        resp.content_type = "application/pdf"
-        return resp
-
-    app = web.Application()
-    app.router.add_get("/", handler)
-    client = await aiohttp_client(app)
-
-    resp = await client.get("/")
-    assert resp.status == 200
-    body = await resp.read()
-    with filepath.open("rb") as f:
-        content = f.read()
-        assert content == body
-    assert resp.headers["Content-Type"] == "application/pdf"
-    assert resp.headers.get("Content-Encoding") is None
-    resp.close()
-    await resp.release()
-    await client.close()
-
-
-async def test_static_file_custom_content_type_compress(aiohttp_client, sender):
-    filepath = pathlib.Path(__file__).parent / "hello.txt"
-
-    async def handler(request):
-        resp = sender(filepath, chunk_size=16)
-        resp.content_type = "application/pdf"
-        return resp
-
-    app = web.Application()
-    app.router.add_get("/", handler)
-    client = await aiohttp_client(app)
-
-    resp = await client.get("/")
-    assert resp.status == 200
-    body = await resp.read()
-    assert b"hello aiohttp\n" == body
-    assert resp.headers["Content-Type"] == "application/pdf"
-    assert resp.headers.get("Content-Encoding") == "gzip"
-    resp.close()
-    await resp.release()
-    await client.close()
-
-
-async def test_static_file_with_gziped_counter_part_enable_compression(
-    aiohttp_client: Any, sender: Any
-):
-    """Test that enable_compression does not double compress when a .gz file is also present."""
-    filepath = pathlib.Path(__file__).parent / "hello.txt"
-
-    async def handler(request):
-        resp = sender(filepath)
-        resp.enable_compression()
-        return resp
-
-    app = web.Application()
-    app.router.add_get("/", handler)
-    client = await aiohttp_client(app)
-
-    resp = await client.get("/")
-    assert resp.status == 200
-    body = await resp.read()
-    assert body == b"hello aiohttp\n"
-    assert resp.headers["Content-Type"] == "text/plain"
-    assert resp.headers.get("Content-Encoding") == "gzip"
-    resp.close()
-    await resp.release()
-    await client.close()
-
-
-async def test_static_file_with_content_encoding(
-    aiohttp_client: Any, sender: Any
+@pytest.mark.parametrize("hello_txt", ["gzip", "br"], indirect=True)
+async def test_static_file_custom_content_type(
+    hello_txt: pathlib.Path, aiohttp_client: Any, sender: Any
 ) -> None:
-    filepath = pathlib.Path(__file__).parent / "hello.txt.gz"
+    """Test that custom type without encoding is returned for encoded request."""
 
     async def handler(request):
-        return sender(filepath)
+        resp = sender(hello_txt, chunk_size=16)
+        resp.content_type = "application/pdf"
+        return resp
 
     app = web.Application()
     app.router.add_get("/", handler)
     client = await aiohttp_client(app)
 
     resp = await client.get("/")
-    assert 200 == resp.status
-    body = await resp.read()
-    assert b"hello aiohttp\n" == body
-    ct = resp.headers["CONTENT-TYPE"]
-    assert "text/plain" == ct
-    encoding = resp.headers["CONTENT-ENCODING"]
-    assert "gzip" == encoding
+    assert resp.status == 200
+    assert resp.headers.get("Content-Encoding") is None
+    assert resp.headers["Content-Type"] == "application/pdf"
+    assert await resp.read() == hello_txt.read_bytes()
+    resp.close()
+    await resp.release()
+    await client.close()
+
+
+@pytest.mark.parametrize(
+    ("accept_encoding", "expect_encoding"),
+    [("gzip, deflate", "gzip"), ("gzip, deflate, br", "br")],
+)
+async def test_static_file_custom_content_type_compress(
+    hello_txt: pathlib.Path,
+    aiohttp_client: Any,
+    sender: Any,
+    accept_encoding: str,
+    expect_encoding: str,
+):
+    """Test that custom type with encoding is returned for unencoded requests."""
+
+    async def handler(request):
+        resp = sender(hello_txt, chunk_size=16)
+        resp.content_type = "application/pdf"
+        return resp
+
+    app = web.Application()
+    app.router.add_get("/", handler)
+    client = await aiohttp_client(app)
+
+    resp = await client.get("/", headers={"Accept-Encoding": accept_encoding})
+    assert resp.status == 200
+    assert resp.headers.get("Content-Encoding") == expect_encoding
+    assert resp.headers["Content-Type"] == "application/pdf"
+    assert await resp.read() == HELLO_AIOHTTP
+    resp.close()
+    await resp.release()
+    await client.close()
+
+
+@pytest.mark.parametrize(
+    ("accept_encoding", "expect_encoding"),
+    [("gzip, deflate", "gzip"), ("gzip, deflate, br", "br")],
+)
+@pytest.mark.parametrize("forced_compression", [None, web.ContentCoding.gzip])
+async def test_static_file_with_encoding_and_enable_compression(
+    hello_txt: pathlib.Path,
+    aiohttp_client: Any,
+    sender: Any,
+    accept_encoding: str,
+    expect_encoding: str,
+    forced_compression: Optional[web.ContentCoding],
+):
+    """Test that enable_compression does not double compress when an encoded file is also present."""
+
+    async def handler(request):
+        resp = sender(hello_txt)
+        resp.enable_compression(forced_compression)
+        return resp
+
+    app = web.Application()
+    app.router.add_get("/", handler)
+    client = await aiohttp_client(app)
+
+    resp = await client.get("/", headers={"Accept-Encoding": accept_encoding})
+    assert resp.status == 200
+    assert resp.headers.get("Content-Encoding") == expect_encoding
+    assert resp.headers["Content-Type"] == "text/plain"
+    assert await resp.read() == HELLO_AIOHTTP
+    resp.close()
+    await resp.release()
+    await client.close()
+
+
+@pytest.mark.parametrize(
+    ("hello_txt", "expect_type"),
+    [
+        ("gzip", "application/gzip"),
+        ("br", "application/x-brotli"),
+        ("bzip2", "application/x-bzip2"),
+    ],
+    indirect=["hello_txt"],
+)
+async def test_static_file_with_content_encoding(
+    hello_txt: pathlib.Path, aiohttp_client: Any, sender: Any, expect_type: str
+) -> None:
+    """Test requesting static compressed files returns the correct content type and encoding."""
+
+    async def handler(request):
+        return sender(hello_txt)
+
+    app = web.Application()
+    app.router.add_get("/", handler)
+    client = await aiohttp_client(app)
+
+    resp = await client.get("/")
+    assert resp.status == 200
+    assert resp.headers.get("Content-Encoding") is None
+    assert resp.headers["Content-Type"] == expect_type
+    assert await resp.read() == hello_txt.read_bytes()
     resp.close()
 
     await resp.release()
@@ -549,15 +604,6 @@ async def test_static_file_directory_traversal_attack(aiohttp_client) -> None:
     await client.close()
 
 
-def test_static_route_path_existence_check() -> None:
-    directory = pathlib.Path(__file__).parent
-    web.StaticResource("/", directory)
-
-    nodirectory = directory / "nonexistent-uPNiOEAg5d"
-    with pytest.raises(ValueError):
-        web.StaticResource("/", nodirectory)
-
-
 async def test_static_file_huge(aiohttp_client, tmp_path) -> None:
     file_path = tmp_path / "huge_data.unknown_mime_type"
 
